/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/






#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include "ABKCommon/pathDelims.h"
#include "hgWDims.h"

void HGraphWDimensions::saveAsNodesNetsWts(const char* baseFileName) const
{
    char* fname  = new char[strlen(baseFileName) + 8];
    strcpy(fname, baseFileName);
    char* extPtr = fname + strlen(baseFileName);

    strcpy(extPtr, ".nets");
    ofstream nets(fname);

    strcpy(extPtr, ".nodes");
    ofstream nodes(fname);

    strcpy(extPtr, ".wts");
    ofstream wts (fname);

    strcpy(extPtr, ".aux");
    ofstream aux (fname);

	//write the nodes file
    nodes<<"UCLA nodes 1.0"<<endl;
    nodes<<TimeStamp()<<User()<<Platform()<<endl;

    nodes<<"NumNodes :     "<<setw(10)<<getNumNodes()<<endl;
    nodes<<"NumTerminals : "<<setw(10)<<getNumTerminals()<<endl;

    itHGFNodeGlobal nIt;
    for(nIt = nodesBegin(); nIt != nodesEnd(); nIt++)
    {
	HGFNode& node   = **nIt;
	unsigned nodeId = node.getIndex();
	Symmetry sym    = _nodeSymmetries[nodeId];

        if (node.getName())	nodes<<setw(10)<<node.getName()<<"  ";
        else   
        {
            if (isTerminal(nodeId)) nodes<<" p"<< nodeId+1 <<" ";
            else                   nodes<<" a"<< nodeId-getNumTerminals()<<" ";
        }

	if(getNodeWidth(nodeId) != 0 || getNodeHeight(nodeId) != 0)
	    nodes<<setw(10)<<getNodeWidth(nodeId)<<" "
                 <<setw(10)<<getNodeHeight(nodeId)<<" ";
	if(isTerminal(nodeId)) nodes<<"terminal ";
	if(sym.rot90 || sym.y || sym.x) nodes << " : "<<sym;
	nodes<<endl;
    }
    nodes.close();


	//write the nets file
    nets<<"UCLA nets 1.0"<<endl;
    nets<<TimeStamp()<<User()<<Platform()<<endl;

    nets<<"NumNets : "<<getNumEdges()<<endl;
    nets<<"NumPins : "<<getNumPins()<<endl;
    
    itHGFEdgeGlobal e;
    for (e=edgesBegin(); e!=edgesEnd(); ++e)
    {
	HGFEdge&	edge   = **e;
	unsigned	edgeId = edge.getIndex();
        const char*     netName=getNetNameByIndex(edgeId);
	
	nets<<"NetDegree : "<<(*e)->getDegree()<<" ";
        if (netName) nets << setw(10) << netName;
        else if (abs(edge.getWeight()-1.0)>1e-10)
                     nets << " net" << edgeId;
        nets << endl;

	unsigned nodeOffset = 0;

        itHGFNodeLocal v;
        for (v=(*e)->srcsBegin(); v!=(*e)->srcsEnd(); v++, nodeOffset++)
        {
            const HGFNode& node=(**v);
            unsigned idx=node.getIndex();
	    nets<<setw(10)<<node.getName()<<" O ";

	    Point pOffset = nodesOnEdgePinOffset(nodeOffset, edgeId);
            double dx=pOffset.x - _nodeWidths [idx]/2,
                   dy=pOffset.y - _nodeHeights[idx]/2;
//	    if(pOffset.x != 0 || pOffset.y != 0)
	    if( abs(dx)>1e-10 || abs(dy)>1e-10)
		nets<<" : "<<dx <<" "  <<dy<<endl;
	    else
	 	nets<<endl;
        }
        for (v=(*e)->srcSnksBegin(); v!=(*e)->srcSnksEnd(); ++v, nodeOffset++)
        {
            const HGFNode& node=(**v);
            unsigned idx=node.getIndex();
	    nets<<setw(10)<<node.getName()<<" B ";

	    Point pOffset = nodesOnEdgePinOffset(nodeOffset, edgeId);
            double dx=pOffset.x - _nodeWidths [idx]/2,
                   dy=pOffset.y - _nodeHeights[idx]/2;
//          if(pOffset.x != 0 || pOffset.y != 0)
	    if( abs(dx)>1e-10 || abs(dy)>1e-10)
		nets<<" : "<<dx <<" "  <<dy<<endl;
	    else nets<<endl;
        }
        for (v=(*e)->snksBegin(); v!=(*e)->snksEnd(); ++v, nodeOffset++)
        {
            const HGFNode& node=(**v);
            unsigned idx=node.getIndex();
	    nets<<setw(10)<<node.getName()<<" I ";

	    Point pOffset = nodesOnEdgePinOffset(nodeOffset, edgeId);
            double dx=pOffset.x - _nodeWidths [idx]/2,
                   dy=pOffset.y - _nodeHeights[idx]/2;

//	    if(pOffset.x != 0 || pOffset.y != 0)
	    if( abs(dx)>1e-10 || abs(dy)>1e-10)
		nets<<" : "<<dx <<" "  <<dy<<endl;
	
	    else nets<<endl;
        }
    }
    nets.close();

		//write the wts file

    wts<<"UCLA wts 1.0"<<endl;
    wts<<TimeStamp()<<User()<<Platform()<<endl;

    unsigned n;
    if(getNumWeights() >= 1)
    {
      for(n=0; n!=getNumNodes(); n++)
      {
	HGFNode& node = *_nodes[n];
	wts<<setw(10)<<node.getName()<<"  ";

        for(unsigned i=0; i!=getNumWeights(); ++i)
	    wts<<" "<<setw(6)<< node.getWeight(i);

        wts<<endl;
      }
    }

    for(n=0; n!=getNumEdges(); ++n)
    {
        const HGFEdge& net=getEdgeByIdx(n);
        if (abs(net.getWeight()-1.0)<1e-10) continue;
        wts << " ";
        const char* netName=getNetNameByIndex(n);
        if (netName) wts << setw(10) << netName ;
        else wts << "net" << n;
        wts << " " << setw(10) << net.getWeight() << endl;
    }

    wts.close();

    aux<<"HGraphWPins : ";

    const char * tmp=NULL;
    tmp=strrchr(baseFileName,pathDelim);
    if (tmp++)
        aux<<tmp<<".nets "<<tmp<<".nodes "<<tmp<<".wts"<<endl;
    else
        aux<<baseFileName<<".nets "<<baseFileName<<".nodes "
	   <<baseFileName<<".wts"<<endl;
    aux.close();

    delete []fname;
}

