/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/






#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <ctype.h>
#include "ABKCommon/pathDelims.h"
#include "Geoms/bbox.h"
#include "RBPlacement.h"

RBPlacement::RBPlacement(const char* auxFileName, const Parameters& params) 
	: _hgWDims(NULL), _populated(false), _cellsNotInRows(0),
	  _placement(0), _isCoreCell(0), _isFixed(0),
	  _params(params), _nonConstThis(this)
{
    ifstream auxFile(auxFileName);
    abkfatal(auxFile, "no aux file found");
    strcpy(origFileName,auxFileName);

//  char dir[255];
    const char *auxFileDirEnd=strrchr(auxFileName,pathDelim);
    if (auxFileDirEnd)
    {
       strncpy(dirName,auxFileName,auxFileDirEnd-auxFileName+1);
       dirName[auxFileDirEnd-auxFileName+1]=0;
    }
    else strcpy(dirName,"");

    bool found=false;
    int  i=0,lineNo=1;
    cout << "Reading " << auxFileName << endl;
    char oneLine[1023];
    char word1[100],word2[100];

    char* newScl  = NULL;
    char* newPl   = NULL;

    while (!found && !auxFile.eof())
    {
        auxFile >> eathash(lineNo) >> word1
                >> noeol(lineNo) >> word2 >> noeol(lineNo);
        abkfatal(!strcmp(word2,":"),
              " Error in aux file: space-separated column expected");
        if(!strcasecmp(word1,"CD"))
        {
            auxFile >> word1;
            auxFile >> needeol(lineNo++);
            if (word1[0]==pathDelimWindows || word1[0] == pathDelimUnix)
                  strcpy(dirName,word1);
            else strcat(dirName,word1);
            char     fDel[2];
            sprintf(fDel,"%c",pathDelim);
            if(   word1[strlen(word1)-1]!=pathDelimWindows
               || word1[0] == pathDelimUnix)
                  strcat(dirName,fDel);
        }
        else if (!strcasecmp(word1,"RowBasedPlacement"))
        {
            found=true;
            auxFile.getline(oneLine,1023);
            unsigned j, len=strlen(oneLine), fileNum=0;
            bool space=true;
            char *fileNames[5] = {NULL,NULL,NULL,NULL,NULL};
            for(j=0; j!=len; j++)
            {
                if(isspace(oneLine[j]))
                {
                    if(!space) oneLine[j]='\0';
                    space=true;
                }
                else if (space)
                {
                    abkfatal(fileNum<5," Too many filenames in AUX file");
                    fileNames[fileNum++]=oneLine+j;
                    space=false;
                }
            }

            for(j=0; j!=fileNum; j++)
            {
                 if (strstr(fileNames[j],".scl"))  newScl =fileNames[j];
                 if (strstr(fileNames[j],".SCL"))  newScl =fileNames[j];
                 if (strstr(fileNames[j],".pl"))   newPl  =fileNames[j];
                 if (strstr(fileNames[j],".PL"))   newPl  =fileNames[j];
            }

            char * tmpStr=NULL;
            unsigned dirLen=strlen(dirName);

            if(newScl!=NULL)
            if(newScl[0]!=pathDelimWindows && newScl[0] != pathDelimUnix )
            {
                  tmpStr=new char[dirLen+strlen(newScl)+1];
                  sprintf(tmpStr,"%s%s",dirName,newScl);
                  char* forSwap = tmpStr;
                  newScl = tmpStr;
                  allFileNamesInAUX.push_back(strdup(newScl));
                  tmpStr = forSwap;
            }

            if(newPl !=NULL)
            if(newPl[0]!=pathDelimWindows && newPl[0] != pathDelimUnix )
            {
                  tmpStr=new char[dirLen+strlen(newPl)+1];
                  sprintf(tmpStr,"%s%s",dirName,newPl);
                  char* forSwap = tmpStr;
                  newPl = tmpStr;
                  allFileNamesInAUX.push_back(strdup(newPl));
                  tmpStr = forSwap;
            }
        }
        if (found)
        while ( ! auxFile.eof() && allFileNamesInAUX.size() < 20)
        {
            auxFile >> word1;
            allFileNamesInAUX.push_back(strdup(word1));
        }
        if ( i++==5) abkfatal(0,"Done");
    }

    HGraphParameters hgParams;
    hgParams.makeAllSrcSnk = false;
    _hgWDims = new HGraphWDimensions(auxFileName, NULL, NULL, hgParams);

    _isCoreCell = bit_vector(_hgWDims->getNumNodes(), true);
    std::fill_n(_isCoreCell.begin(), _hgWDims->getNumTerminals(), false);

    _isFixed    = bit_vector(_hgWDims->getNumNodes(), false);
    std::fill_n(_isFixed.begin(),    _hgWDims->getNumTerminals(), true);

    _isInSubRow = bit_vector(_hgWDims->getNumNodes(), false);
    std::fill_n(_isInSubRow.begin(), _hgWDims->getNumTerminals(), true);

    _cellCoords = vector<RBCellCoord>(_hgWDims->getNumNodes());

    constructRows(newScl);

    if(newPl) 
    {
	//have to parse the placement file here, as it may have
	//node names, which only the HGraph knows.
	_placement = PlacementWOrient(_hgWDims->getNumNodes(),
                     Point(DBL_MAX,DBL_MAX));

	ifstream plFile(newPl);
        abkfatal2(plFile, " Could not open ",newPl);
        cout << " Reading " << newPl << " ... " << endl;

	char     nodeName[256];
	Point    pt;
        char     oriStr[10];
        int lineno = 0;

	plFile>>needcaseword("UCLA")>>needcaseword("pl")
	      >>needword("1.0")>>skiptoeol; 
	plFile>>eathash(lineno);

	while(!plFile.eof())
        {
	    plFile>>eathash(lineno)>>nodeName>>pt.x>>pt.y>>eatblank;
	    unsigned nodeId = _hgWDims->getNodeByName(nodeName).getIndex();

	    _placement[nodeId] = pt;
	    if(plFile.peek() == ':')
	    {
		plFile>>needword(":")>>oriStr;
		_placement.setOrient(nodeId, Orientation(oriStr));
	    }
	    plFile>>skiptoeol>>eathash(lineNo);
	}	
    }
    else
	_placement = PlacementWOrient(_hgWDims->getNumNodes(),
                     Point(DBL_MAX,DBL_MAX));
    cout << "Finished reading " << auxFileName << " and referenced files"<<endl;

    setCellsInRows();
    populateCC();

    if(newScl) delete []newScl;
    if(newPl)  delete []newPl;
}

void RBPlacement::constructRows(const char* sclFileName)
{
    ifstream sclFile(sclFileName);
    abkfatal2(sclFile, " Could not open ",sclFileName);
    cout << " Reading " << sclFileName << " ... " << endl;

    int lineNo = 0;
    unsigned numRows = 0;
    char oneLine[100];


    sclFile >> eathash(lineNo) >> needcaseword("UCLA",lineNo)
            >> needcaseword("scl",lineNo) >> skiptoeol;
    lineNo++;
    sclFile >> eathash(lineNo);

    sclFile>>needcaseword("NumRows")>>needword(":")>>numRows>>skiptoeol;
    _coreRows.reserve(numRows*2);

    double   yCoord=-1.23456;
    double   height=-1.23456;
    double   width=-1.23456;
    double   spacing=-1.23456;
    char     oriStr[10];
    char     symStr[10];
    double   xStart=-1.23456;
    unsigned numSites=UINT_MAX;

    lineNo = 3;

    for(unsigned rNum = 0; rNum < numRows; rNum++)
    {
	strcpy(oriStr," ");
	strcpy(symStr," ");
	spacing = DBL_MAX;


	sclFile>>needcaseword("CoreRow")>>needcaseword("Horizontal")>>skiptoeol;
	lineNo++;

	sclFile >> needcaseword("Coordinate")  >> needword(":", lineNo) 
		>> isnumber(lineNo)>>yCoord>>skiptoeol;
	lineNo++;

	sclFile >> eathash(lineNo);

	if(sclFile.peek() == 'H' || sclFile.peek() == 'h')
	{
	  sclFile >> needcaseword("Height")  >> needword(":", lineNo) 
		>> isnumber(lineNo)>>height>>needeol();
	  lineNo++;
    	}

        sclFile >> eathash(lineNo) >> oneLine;
	while(strcasecmp(oneLine, "SubrowOrigin"))
	{
	  if(!strcasecmp(oneLine, "Sitewidth"))
		sclFile >> needword(":", lineNo) >> isnumber(lineNo)
		        >> width >> skiptoeol;
          else if(!strcasecmp(oneLine, "Sitespacing"))
		sclFile >> needword(":", lineNo) >> isnumber(lineNo)
		        >> spacing >> skiptoeol;
	  else if(!strcasecmp(oneLine, "Siteorient"))
		sclFile >> needword(":", lineNo) >> oriStr>>skiptoeol;
	  else if(!strcasecmp(oneLine, "Sitesymmetry"))
		sclFile >> needword(":", lineNo) >> symStr>>skiptoeol;
          else abkfatal3(0,"Unexpected word : ",oneLine,
  "\n(expected one of SubRowOrigin, Sitewidth, Siteorient, Sitesymmetry)\n");
	  lineNo++;
          sclFile >> eathash(lineNo) >> oneLine;
	}

	spacing = (spacing==DBL_MAX) ? width : spacing;

	_coreRows.push_back(RBPCoreRow(yCoord, Orient(oriStr),
		RBPSite(height, width, Symmetry(symStr)), _placement, spacing));

        RBPCoreRow& curCoreRow = _coreRows.back();

	while(!strcasecmp(oneLine, "SubrowOrigin"))
	{
	    sclFile>>needword(":", lineNo)>>xStart
		>>needcaseword("Numsites")>>needword(":")>>numSites
		>>skiptoeol;

	    double xEnd = (numSites-1)*spacing + width + xStart;
	    curCoreRow.appendNewSubRow(xStart, xEnd);

	    sclFile >> oneLine;
	    lineNo++;
	}
        abkfatal2(!strcasecmp(oneLine, "End") || sclFile.eof(),
		"in SCL file, expected End, got ", oneLine);	
	sclFile>>eatblank;

	lineNo++;
    }

}

void RBPlacement::saveAsNodesNetsWts(const char* baseFileName) const
{
    _hgWDims->saveAsNodesNetsWts(baseFileName);
    saveAsSCL(baseFileName);

    char *plName = new char[strlen(baseFileName) + 4];
    strcpy(plName, baseFileName);
    strcpy(plName+strlen(baseFileName), ".pl");

    ofstream plOut(plName);
    plOut<<"UCLA pl 1.0"<<endl;
    plOut<<TimeStamp() << User() << Platform() << endl << endl;

    for(unsigned i=0; i<_hgWDims->getNumNodes(); i++)
    {
      if (_placement[i].x > 1e308 || _placement[i].x > 1e308) continue;

      plOut << _hgWDims->getNodeByIdx(i).getName() <<"  "
	    << setw(10) << _placement[i].x << "  "
	    << setw(10) << _placement[i].y << " : "
	    << _placement.getOrient(i)<<endl;
    }
    plOut.close();
   
    char *auxName = new char[strlen(baseFileName) + 5];
    strcpy(auxName, baseFileName);
    strcpy(auxName+strlen(baseFileName), ".aux");

    ofstream auxFile(auxName);

    auxFile<<"RowBasedPlacement : ";

    const char * tmp=NULL;
    tmp=strrchr(baseFileName,pathDelim);
    if (tmp++)
        auxFile<<tmp<<".nodes "
    	       <<tmp<<".nets "
	       <<tmp<<".wts "
	       <<tmp<<".pl "
	       <<tmp<<".scl"<<endl;
    else
        auxFile<<baseFileName<<".nodes "
    	       <<baseFileName<<".nets "
	       <<baseFileName<<".wts "
	       <<baseFileName<<".pl "
	       <<baseFileName<<".scl"<<endl;

    auxFile.close();
    delete []auxName;
}

void RBPlacement::saveAsNetDAre(const char* baseFileName) const
{
    _hgWDims->saveAsNetDAre(baseFileName);
    saveAsSCL(baseFileName);

    char *plName = new char[strlen(baseFileName) + 4];
    strcpy(plName, baseFileName);
    strcpy(plName+strlen(baseFileName), ".pl");

    _placement.save(plName, _hgWDims->getNumTerminals());
   
    char *auxName = new char[strlen(baseFileName) + 5];
    strcpy(auxName, baseFileName);
    strcpy(auxName+strlen(baseFileName), ".aux");

    ofstream auxFile(auxName);

    auxFile<<"RowBasedPlacement : ";
    auxFile<<baseFileName<<".netD "
	   <<baseFileName<<".areM "
	   <<baseFileName<<".pl "
	   <<baseFileName<<".scl"<<endl;

    auxFile.close();
    delete []auxName;
}

void RBPlacement::saveAsSCL(const char* baseFileName) const
{
    char* fname = new char[strlen(baseFileName)+4];
    strcpy(fname, baseFileName);
    strcat(fname, ".scl");
    ofstream sclFile(fname);

    sclFile<<"UCLA scl 1.0"<<endl;
    sclFile<< TimeStamp() << User() << Platform() << endl << endl;

    sclFile<<"Numrows : "<<_coreRows.size()<<endl<<endl;

    unsigned r;
    for(r = 0; r < _coreRows.size(); r++)
        _coreRows[r].save(sclFile);
    sclFile.close();
}

void RBPlacement::saveAsSpatialConstraints (const char* constFileName) const
{
    char *tempFileName = new char[strlen(constFileName)+40];
    strcpy(tempFileName, constFileName);
    strcat(tempFileName, ".simplelc");

    ofstream constFile(constFileName);

    vector<double> validYLocs;
    BBox layoutBBox;

    for(itRBPCoreRow cr = coreRowsBegin(); cr != coreRowsEnd(); cr++)
    {
	layoutBBox += Point((cr)->getXStart(), (cr)->getYCoord());
	layoutBBox += Point((cr)->getXEnd(),  
				(cr)->getYCoord()+(cr)->getHeight());

	validYLocs.push_back((cr)->getYCoord());
    }

    constFile<<"Layout BBox ("<<layoutBBox.xMin<<", "<<layoutBBox.yMin
	     <<", "<<layoutBBox.xMax<<", "<<layoutBBox.yMax<<")"<<endl;

    constFile<<"validY      ("<<validYLocs[0];
    for(unsigned y = 1; y < validYLocs.size(); y++)
	constFile<<", "<<validYLocs[y];
    constFile<<")"<<endl;

    constFile.close();
    delete [] tempFileName;
}

const char *RBPlacement::getAuxName()        const
{
   const char *left=strrchr(origFileName,pathDelim);
   if (left==NULL) left=origFileName;
   else            left++;;


   const char *right=strrchr(left,'.');
   if (right==NULL) return left;

   static char buf[256];
   strncpy(buf,left,right-left);
   if (right-left==0) strcpy(buf,"noname");
   return buf;
}
