/**************************************************************************
***    
*** Copyright (c) 1995-2000 Regents of the University of California,
***               Andrew E. Caldwell, Andrew B. Kahng and Igor L. Markov
***
***  Contact author(s): abk@cs.ucla.edu, imarkov@cs.ucla.edu
***  Affiliation:   UCLA, Computer Science Department,
***                 Los Angeles, CA 90095-1596 USA
***
***  Permission is hereby granted, free of charge, to any person obtaining 
***  a copy of this software and associated documentation files (the
***  "Software"), to deal in the Software without restriction, including
***  without limitation 
***  the rights to use, copy, modify, merge, publish, distribute, sublicense, 
***  and/or sell copies of the Software, and to permit persons to whom the 
***  Software is furnished to do so, subject to the following conditions:
***
***  The above copyright notice and this permission notice shall be included
***  in all copies or substantial portions of the Software.
***
*** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
*** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
*** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
*** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
*** CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
*** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
*** THE USE OR OTHER DEALINGS IN THE SOFTWARE.
***
***
***************************************************************************/



//by sadya to generate gnuplots for nets
#ifdef _MSC_VER
#pragma warning(disable:4786)
#endif

#include <ctype.h>
#include "ABKCommon/pathDelims.h"
#include "Geoms/bbox.h"
#include "RBPlacement.h"

void RBPlacement::saveAsPlot(int argc, const char* argv[],
			     const char* baseFileName) const
{
    char* fname = new char[strlen(baseFileName)+4];
    strcpy(fname, baseFileName);
    strcat(fname, ".gpl");
    ofstream gpFile(fname);
    unsigned i=0;
    double xCentre=0,yCentre=0;
    double rowHeight=0;
    double nodeWidth=0;

    vector<double> xcoord;
    vector<double> ycoord;

    StringParam   plotNets("plotNets", argc, argv);
    StringParam   plotNodes("plotNodes", argc, argv);
    StringParam   plotNodesAndNets("plot", argc, argv);
    StringParam   plotNodesWNames("plotNodesWNames", argc, argv);
    StringParam   plotNodesAndNetsWNames("plotWNames", argc, argv);

    DoubleParam xmin("xmin",argc,argv);
    DoubleParam xmax("xmax",argc,argv);
    DoubleParam ymin("ymin",argc,argv);
    DoubleParam ymax("ymax",argc,argv);

    Point minRange;
    Point maxRange;

    minRange.x  = ( xmin.found() ?  xmin:-DBL_MAX);
    maxRange.x  = ( xmax.found() ?  xmax:DBL_MAX);
    minRange.y  = ( ymin.found() ?  ymin:-DBL_MAX);
    maxRange.y  = ( ymax.found() ?  ymax:DBL_MAX);

    gpFile<<"#UCLA gpl 1.0"<<endl;
    gpFile<<"#"<< TimeStamp() << User() << Platform() << endl << endl;
    gpFile << "set title 'CELLS "<<endl<<endl;
    gpFile << "# set terminal postscript eps color 24"<<endl;
    gpFile << "# set output \"out.eps\""<<endl<<endl<<endl;

    //plot names of cells first if required
    if(plotNodesWNames.found() || plotNodesAndNetsWNames.found())
      {
	for(i=0; i<_hgWDims->getNumNodes(); i++)
	  {
	    gpFile <<"set label \""<<_hgWDims->getNodeByIdx(i).getName() 
		 <<"\" at "<< _placement[i].x+_hgWDims->getNodeWidth(i)/2
		  <<" , "<<_placement[i].y+_hgWDims->getNodeHeight(i)/2<<endl;
	  }
      } 

    gpFile << endl<<endl;

    //get the customized plot command for each option.
    gpFile << " plot[";
    if(xmin.found())        gpFile<<minRange.x;
    gpFile<<":";
    if(xmax.found())        gpFile<<maxRange.x;
    gpFile<<"][";
    if(ymin.found())        gpFile<<minRange.y;
    gpFile<<":";
    if(ymax.found())        gpFile<<maxRange.y;
    gpFile<<"]";

    if(plotNets.found())
      gpFile<<" '-' w l ";
    if(plotNodes.found())
       gpFile << " '-' w l 3 ";
    if(plotNodesAndNets.found())
       gpFile << " '-' w l, '-' w l 3 ";
    if(plotNodesWNames.found())
       gpFile << " '-' w l 3 ";
    if(plotNodesAndNetsWNames.found())
       gpFile << " '-' w l, '-' w l 3 ";
    gpFile << endl<<endl;


 if(plotNets.found() || plotNodesAndNets.found()
    || plotNodesAndNetsWNames.found())
 {
    // the hyperedges info goes inside

    itHGFEdgeGlobal e;
    const char* cellName ;
    unsigned netDegree = 0;
    Point lastRowPoint;
    for (e=_hgWDims->edgesBegin(); e!=_hgWDims->edgesEnd(); ++e)
    {
        netDegree=(*e)->getDegree();
	gpFile<<"#NetDegree : "<<netDegree<<endl;
	
	xcoord.resize(netDegree);
	ycoord.resize(netDegree);
	HGFEdge&	edge   = **e;
	unsigned edgeId =  edge.getIndex();
	unsigned nodeOffset = 0;

        itHGFNodeLocal v;
	unsigned j=0;
    // for all nodes
        for (v=(*e)->nodesBegin(),j=0; v!=(*e)->nodesEnd(); v++, nodeOffset++)
        {
	    cellName = (*v)->getName();
	    gpFile<<"#"<<setw(10)<<cellName<<" O "<<endl;
	    HGFNode& node = _hgWDims->getNodeByName(cellName);
	    i=node.getIndex();
	    Point tempRowPoint(_placement[i].x,_placement[i].y);
	    lastRowPoint=tempRowPoint;

	    Point pOffset = _hgWDims->nodesOnEdgePinOffset(nodeOffset, edgeId);
            ycoord[j]= _placement[i].y+pOffset.y;
	    xcoord[j] = _placement[i].x+pOffset.x;
	    j++;
	    rowHeight=_hgWDims->getNodeHeight(i);
	    nodeWidth=_hgWDims->getNodeWidth(i);
        }

    unsigned withinRange=0;

    for(i=0; i<netDegree;++i)
      {
	if((xcoord[i]>minRange.x && xcoord[i]<maxRange.x)
         &&(ycoord[i]>minRange.y && ycoord[i]<maxRange.y))
	  withinRange=1;   //atleast 1 cell of Hedge is within range
	                   //so plot the whole net
      }


    if(withinRange == 1)
      {
	unsigned status=0;

        for(i=1; i<netDegree;++i)
	  {
	    if(ycoord[0]!=ycoord[i]) 
	      {
	        status = 1;   //all cells of Hedge are not in single row
	        break;
	      }
	  }

	 if(status==1)
	   {
	     for(i=0,xCentre=0; i<netDegree;++i)
	       {
		 xCentre += xcoord[i];
	       }
	     xCentre /= netDegree;

	     for(i=0,yCentre=0; i<netDegree;++i)
	       {
		 yCentre += ycoord[i];
	       }
	     yCentre /= netDegree;
	   }
	 else
	   {
	     const RBPCoreRow* rowName;
	     rowName = findCoreRow(lastRowPoint);
	     double rowLength = fabs(rowName->getXEnd()-rowName->getXStart());

	     double maxXcoord=0;
	     double minXcoord=0;

	     minXcoord=*min_element(xcoord.begin(),xcoord.end());
	     maxXcoord=*max_element(xcoord.begin(),xcoord.end());
	     double ydifference = maxXcoord-minXcoord;
	     double rd = RandomDouble(-1,1);
	     int starOrient=0;

	     starOrient = ((rd>0) ? 1:-1);//get orientation of star point

	     rd = RandomDouble(-1,1);
	     ydifference = (1-exp(0-ydifference*ydifference/
			   (rowLength)))*rowHeight*0.4*starOrient;
	     yCentre = ycoord[0]+ydifference;
	     xCentre = (maxXcoord+minXcoord)/2+rd*nodeWidth*0.75;
	   }

	 if(status == 0 || netDegree != 2)
	   {
	     if(abs(xCentre)<1e+306 && abs(yCentre)<1e+306)
	       {
		 gpFile << setw(10) << xCentre << "  "
			<< setw(10) << yCentre+0.025*rowHeight << "  "<<endl;
		 gpFile << setw(10) << xCentre+0.025*rowHeight << "  "
			<< setw(10) << yCentre << "  "<<endl;
		 gpFile << setw(10) << xCentre << "  "
			<< setw(10) << yCentre-0.025*rowHeight << "  "<<endl;
		 gpFile << setw(10) << xCentre-0.025*rowHeight << "  "
			<< setw(10) << yCentre << "  "<<endl;
		 gpFile << setw(10) << xCentre << "  "
		     << setw(10) << yCentre+0.025*rowHeight <<"  "<<endl<<endl;
	       }
	   }
      
	  for(i=0;i<netDegree;++i)
	    {
	    if(abs(xcoord[i])<DBL_MAX && abs(ycoord[i])<DBL_MAX &&
	       abs(xCentre)<DBL_MAX && abs(yCentre)<DBL_MAX)
	      {
	      gpFile << setw(10) << xCentre << "  "
	             << setw(10) << yCentre << "  "<<endl;
	      gpFile << setw(10) << xcoord[i] << "  "
	             << setw(10) << ycoord[i] << "  "<<endl;
	      gpFile << setw(10) << xCentre << "  "
	             << setw(10) << yCentre << "  "<<endl<<endl;
	      }
	    }
      }
    }
    gpFile <<"EOF"<<endl;
 }

//added by sadya to plot nodes
//plot the actual nodes now

 if(plotNodes.found() || plotNodesAndNets.found() || plotNodesWNames.found() ||
 plotNodesAndNetsWNames.found())
 {
    double cellWidth=0;
    double cellHeight=0;

    for(i=0; i<_hgWDims->getNumNodes(); i++)
    {
      cellWidth=_hgWDims->getNodeWidth(i);
      cellHeight=_hgWDims->getNodeHeight(i)*0.9;
      double offset = 0;

      offset = ((cellWidth > cellHeight) ? 0.1*cellHeight:0.1*cellWidth);
      gpFile <<"# Cell Name "<<_hgWDims->getNodeByIdx(i).getName()<<endl;
       
    if(abs(_placement[i].x) < 1e+306 && abs(_placement[i].y) < 1e+306)
     {
      gpFile << setw(10) << _placement[i].x << "  "
	    << setw(10) << _placement[i].y+offset << "   "
	    <<endl;
      gpFile << setw(10) << _placement[i].x+offset << "  "
	    << setw(10) << _placement[i].y << "   "
	    <<endl;


      gpFile << setw(10) << _placement[i].x+cellWidth-offset << "  "
	    << setw(10) << _placement[i].y << "   "
	    <<endl;
      gpFile << setw(10) << _placement[i].x+cellWidth << "  "
	    << setw(10) << _placement[i].y+offset << "   "
	    <<endl;


      gpFile << setw(10) << _placement[i].x+cellWidth << "  "
	    << setw(10) << _placement[i].y+cellHeight-offset << "   "
	    <<endl;
      gpFile << setw(10) << _placement[i].x+cellWidth-offset << "  "
	    << setw(10) << _placement[i].y+cellHeight << "   "
	    <<endl;


      gpFile << setw(10) << _placement[i].x+offset << "  "
	    << setw(10) << _placement[i].y+cellHeight << "   "
	    <<endl;
      gpFile << setw(10) << _placement[i].x << "  "
	    << setw(10) << _placement[i].y+cellHeight-offset << "   "
	    <<endl;


      gpFile << setw(10) << _placement[i].x << "  "
	    << setw(10) << _placement[i].y+offset << "   "
	    <<endl;
      gpFile <<endl<<endl;
     }
    }
    gpFile << "EOF"<<endl;
 }

   gpFile << "pause -1 'Press any key' "<<endl;
   gpFile.close();
}
