/**************************************************************************
***
***    Copyright (c) 1995-1997 by ABKGroup, UCLA VLSI CAD Laboratory,
***    UCLA Computer Science Department, Los Angeles, CA 90095-1596 USA,
***    and by the Regents of the University of California.
***    All rights reserved.
***
***    No portion of this file may be used, copied, or transmitted for
***    any purpose, including any research or commercial purpose,
***    without the express written consent of Professor Andrew B. Kahng,
***    UCLA Computer Science Department, 3713 Boelter Hall, Los Angeles,
***    CA 90095-1596 USA.   Any copy made of this file must include this
***    notice.    ABKGroup and the Regents of the University of California
***    retain title to all copies of this file and/or its compiled
***    derivatives.
***
***************************************************************************/

// Created: April 1, 1998 by I.Markov by splitting A.Caldwell's capoPlacer.cxx

//CHANGES
//970923 aec moving to new Capo structure: major changes include removal pf
//		padBlock, addition of a persistent slice, move away from
//		the 'changeNetlist' 'changeLayout' design into a unified
//		'refineBlock' structure.
//971113 aec added class CapoPlacer::Parameters
//971118 aec added SA multi-start and multi-level params to 
//			CapoPlacer::Paremeters
//980215 aec moved parameter code to capoParams.cxx
//980305 ilm removed fixedConstraint handling for Partitioner::Parameters
//       ilm split capoPlacer.cxx into this file, splitBlock.cxx and the rest
//       ilm introduced "CapoPlacer::splitBlock()"
//990319 ilm refineBlock() now returns bool "this block has children that
//           need to be partitioned/placed"

#include "abkcommon.h"
#include "placement.h"
#include "capoPlacer.h"
#include "capoBlock.h"
#include "partProbForCapo.h"
#include "sRowBBPlacer.h"
#include "splitSmall.h"
#include "splitLarge.h"
#include "laSplitter.h"
#include "splitRow.h"
#include "kPlSplitter.h"

bool CapoPlacer::refineBlock(CapoBlock&          block,
			     vector<CapoBlock*>& newLayout, 
			     unsigned            layerNum, 
			     AllowedPartDir      partDir,
			     unsigned            minNumCellsToPart)
{
    static_cast<void>(layerNum);
    unsigned numLeaves = block.getNumCells();   

    //if(_params.trackRentNums && numLeaves >= 10)
    //	storeRentNumbers(block);
 
    if(numLeaves == 1)
    {	
	unsigned nodeId = *(block.cellIdsBegin());

	_placement[nodeId] = 
		Point(block.getBBox().xMin,(*block.rowsBegin())->getYCoord());
	_placed[nodeId] = true;
	_cellToBlockMap[nodeId] = NULL;

	if(_params.verb.forActions > 0) cerr<<"1";

	if(_params.replaceSmallBlocks == AtEveryLayer)
            newLayout.push_back(&block);
	else
            _placedBlocks.push_back(&block);

        return false;
        // without deleting the block
    }
    else if(block.getNumRows() == 1 && numLeaves <= _params.smallPlaceThreshold)
    {
 	abkwarn(block._startOffsets[0].first == block._endOffsets[0].first,
			"smallPlacer called on a block w/ subrows");

	const RBPCoreRow& onlyRow = **block.rowsBegin();
	double spacing = onlyRow.getSpacing();
        double yCoord  = onlyRow.getYCoord();

        SmallPlacementRow splRow(block.getBBox(), spacing);

        SmallPlacementProblem problem(
			        _hgraphOfNetlist, _placement, _placed,
  			        block.getCellIds(), splRow);

        SmallPlParams splParams = _params.smplParams;
        splParams.verb.forMajStats /= 10;
        splParams.verb.forActions  /= 10;
        splParams.verb.forSysRes   /= 10;
	BranchSmallPlacer placer(problem, splParams);

	const Placement& soln = problem.getBestSoln().placement;
	abkassert(soln.getSize() == block.getNumCells(), "undersized placment");

	double leftMostSite = 
			onlyRow[block._startOffsets[0].first].getXStart();

  	for(unsigned n = 0; n < block.getNumCells(); n++)
	{
	    _placed [block._cellIds[n]] = true;
	    _cellToBlockMap[block._cellIds[n]] = NULL;

	    double locOffset = soln[n].x - leftMostSite;
	    unsigned numSites;
	    if(locOffset <= 0) numSites = 0;
	    else   numSites = static_cast<unsigned>(floor(locOffset/spacing));

	    _placement[block._cellIds[n]] = 
			Point(numSites*spacing+leftMostSite, yCoord);
	}

	if(_params.verb.forActions > 0) cerr<<"-";

	block.setIndexToNextIndex();

	if(_params.replaceSmallBlocks == AtEveryLayer)
            newLayout.push_back(&block);
	else
            _placedBlocks.push_back(&block);

        return false;
        // without deleting the block
    } 
    else
    {
	unsigned numRows = block.getNumRows();
	Verbosity sVerb = _params.verb;
	sVerb.forActions  /= 2;
	sVerb.forMajStats /= 2;
	sVerb.forSysRes   /= 2;

	if(numLeaves < minNumCellsToPart)
	{
	    if(_params.verb.forActions > 4)
		cout<<"Block below current min to split...skipping"<<endl;
	    newLayout.push_back(&block);
	    cerr<<"o";
	    return false;
        }

	BaseBlockSplitter* splitter;
	vector<CapoBlock*> newBlocks;
	updateInfoAboutBlock(&block);

	unsigned bFactor = _params.splitterParams.numMLSets;
	unsigned lAhead  = _params.lookAhead;
	if(numLeaves < 500) lAhead = 0;

	if(_layerNum == 0)
		bFactor *= 3;
	else if(_layerNum == 1)
		bFactor *= 2;

	if(numRows > 1 && 
	   (numLeaves < 15*numRows || block.getHeight() > block.getWidth()))
	{		//Horizontal partitioning
	    if(partDir == VOnly)
	    {
		if(_params.verb.forActions > 4)
		    cout<<"Block wants to split the other way...skipping"<<endl;
		newLayout.push_back(&block);
	    	cerr<<"x";
	        return false;
	    }

	    if(numLeaves <= 200)
		splitter = new SplitSmallBlockHorizontally(block,*this, sVerb);
							
	    else
	    {
		if(lAhead > 0)
	            splitter = new LookAheadSplitter(block,*this, bFactor,
					lAhead, sVerb);
		if(partDir == HAndV || !_params.allowRowSplits)	
		    splitter = new SplitLargeBlockHorizontally(block,*this, 
					_params.splitterParams);
		else
		{
				//collect a row of blocks
		    vector<CapoBlock*> rowOfBlocks;
		    rowOfBlocks.push_back(&block);
		    double rowTop = rowOfBlocks[0]->getBBox().yMax;
		    double rowBot = rowOfBlocks[0]->getBBox().yMin;

		    CapoBlock*  adjBlk = &block;

		    while(adjBlk->_leftNeighbors.size() == 1)
		    {
			CapoBlock* lNbor = adjBlk->_leftNeighbors[0];

			if(lNbor->getNumCells() < 100 ||
			   lNbor->getNumCells() < minNumCellsToPart ||
			   !lNbor->canSplitBlock()) 
				break;

			if((lNbor->getBBox().yMin != rowBot)  ||
			   (lNbor->getBBox().yMax != rowTop))   
				break;
		    
			adjBlk = lNbor;
		        rowOfBlocks.push_back(adjBlk);
		    }

		    adjBlk = &block;
		    while(adjBlk->_rightNeighbors.size() == 1)
		    {
			CapoBlock* rNbor = adjBlk->_rightNeighbors[0];

			if(rNbor->getNumCells() < 100 ||
			   rNbor->getNumCells() < minNumCellsToPart ||
			   !rNbor->canSplitBlock()) 
				break;

			if((rNbor->getBBox().yMin != rowBot)  ||
			   (rNbor->getBBox().yMax != rowTop))   
				break;
		    
			adjBlk = rNbor;
		        rowOfBlocks.push_back(adjBlk);
		    }


		    if(rowOfBlocks.size() > 1)	
		    {
	            	splitter = new SplitRowOfBlocks(rowOfBlocks,
		    				*this, _params.splitterParams);
			if(static_cast<SplitRowOfBlocks*>(splitter)->
				wasGoodSplit())
		 	{
			    for(unsigned sb = 0; sb < rowOfBlocks.size(); sb++)
				rowOfBlocks[sb]->_canSplitBlock = false;
			}
			else
			{
			    cout<<"Failed Row Split"<<endl;	
			    delete splitter;
		            splitter = new SplitLargeBlockHorizontally(
					block,*this, _params.splitterParams);
			}
		    }
		    else
		        splitter = new SplitLargeBlockHorizontally(
					block,*this, _params.splitterParams);
		}
	    }
	}
	else
	{		//Vertical partitioning
	    if(partDir == HOnly)
	    {
		if(_params.verb.forActions > 4)
		    cout<<"Block wants to split the other way...skipping"<<endl;
		newLayout.push_back(&block);
	    	cerr<<"x";
	        return false;
	    }

	    if(numLeaves < 200)
		splitter = new SplitSmallBlockVertically(block,*this,sVerb);
	    else
	    {
		if(lAhead > 0)
	            splitter = new LookAheadSplitter(block,*this,
						bFactor,lAhead,
						sVerb);
		else
		{
		    if(_params.useKPl & _layerNum > 1)
			splitter = new KPlaceSplitter(
					block, *this, _kPlacement, 
					_params.splitterParams);
		    else
		    	splitter = new SplitLargeBlockVertically(
					block, *this, _params.splitterParams);
		}
	    }
	}

	newBlocks = splitter->getNewBlocks();
	delete splitter;

	if(numLeaves <= _params.smallPartThreshold)
	    cerr<<"*";
	else
	    cerr<<".";

	for(unsigned b = 0; b < newBlocks.size(); b++)
	{
	    newLayout.push_back(newBlocks[b]);
	    updateInfoAboutBlock(newBlocks[b]);
	}

	block._canSplitBlock = false;
        return true;
    }
}

